#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>

#include "sbox8.h"

#define SNAMEC  "gl_S"
#define SNAMELD "gl_ddt"
#define SNAMELL "gl_lat"
#define SNAMELA "gl_anf"

#define S4NAMEC  "rand4_S"
#define S4NAMELD "rand4_ddt"
#define S4NAMELL "rand4_lat"
#define S4NAMELA "rand4_anf"

#define Scalar8(a,b) (__builtin_popcount(((a) & (b)) & 0xff) & 1)
#define Scalar4(a,b) (__builtin_popcount(((a) & (b)) & 0xf) & 1)

uint8_t S8i[256] = {0};

uint8_t piccolo_s[16] = {0xe, 0x4, 0xb, 0x2, 0x3, 0x8, 0x0, 0x9, 0x1, 0xa, 0x7, 0xf, 0x6, 0xc, 0x5, 0xd};
uint8_t os1[16] = {0,1,2,3,4,6,8,0xA,5,0xB,0xC,0xF,7,0xD,9,0xE};
uint8_t rand4[16] = {0x04, 0x08, 0x0a, 0x0d, 0x0e, 0x02, 0x0c, 0x01, 0x06, 0x00, 0x03, 0x0b, 0x07, 0x0f, 0x09, 0x05};
uint8_t whirlpool_s0[16] = {0x1, 0xb, 0x9, 0xc, 0xd, 0x6, 0xf, 0x3, 0xe, 0x8, 0x7, 0x4, 0xa, 0x2, 0x5, 0x0};
uint8_t whirlpool_s1[16] = {0x7, 0xc, 0xb, 0xd, 0xe, 0x4, 0x9, 0xf, 0x6, 0x3, 0x8, 0xa, 0x2, 0x5, 0x1, 0x0};
uint8_t hb2_s0[16] = {0x7, 0xC, 0xE, 0x9, 0x2, 0x1, 0x5, 0xF, 0xB, 0x6, 0xD, 0x0, 0x4, 0x8, 0xA, 0x3};

#define S8inv(x) S8i[(x)]

#define S4(x) os1[x]

uint8_t feistel3(uint8_t x)
{
	uint8_t hi = (x >> 4) & 0xF;
	uint8_t lo = x & 0xF;
	uint8_t tmp;

	tmp = hi;
	hi = lo;
	lo = whirlpool_s0[lo] ^ tmp;

	tmp = hi;
	hi = lo;
	lo = whirlpool_s1[lo] ^ tmp;

	tmp = hi;
	hi = lo;
	lo = hb2_s0[lo] ^ tmp;

	return ((hi << 4) ^ lo);

}

uint8_t misty3(uint8_t x)
{
	uint8_t hi = (x >> 4) & 0xF;
	uint8_t lo = x & 0xF;
	uint8_t tmp;

	tmp = hi;
	hi = lo;
	lo = whirlpool_s0[tmp] ^ lo;

	tmp = hi;
	hi = lo;
	lo = whirlpool_s1[tmp] ^ lo;

	tmp = hi;
	hi = lo;
	lo = hb2_s0[tmp] ^ lo;

	return ((hi << 4) ^ lo);

}

uint8_t lm(uint8_t x)
{
	uint8_t hi = (x >> 4) & 0xF;
	uint8_t lo = x & 0xF;
	uint8_t tmp;

	tmp = whirlpool_s0[hi ^lo];

	hi = whirlpool_s1[hi ^tmp];
	lo = hb2_s0[lo ^ tmp];

	return ((hi << 4) ^ lo);

}

uint8_t gl(uint8_t x)
{
	uint8_t r0 = 0x8F;
	uint8_t r1 = 0xB7;
	uint8_t r2 = 0xE3;
	uint8_t r3 = 0xF1;
	uint8_t r4 = 0xF8;
	uint8_t r5 = 0x7B;
	uint8_t r6 = 0x3E;
	uint8_t r7 = 0x1F;

	uint8_t y0,y1,y2,y3,y4,y5,y6,y7;

	y0 = Scalar8(r0,x); 
	y1 = Scalar8(r1,x); 
	y2 = Scalar8(r2,x); 
	y3 = Scalar8(r3,x); 
	y4 = Scalar8(r4,x); 
	y5 = Scalar8(r5,x); 
	y6 = Scalar8(r6,x); 
	y7 = Scalar8(r7,x);
	
	return (y0 ^
			(y1 << 1) ^
			(y2 << 2) ^
			(y3 << 3) ^
			(y4 << 4) ^
			(y5 << 5) ^
			(y6 << 6) ^
			(y7 << 7));
}


// side effect: computes the inverse
int check_invertible(void)
{
	int im[256] = {0};

	for (int i = 0; i < 256; i++)
	{
		if (im[S8(i)])
		{
			printf("Not invertible\n");
			exit(1);
			return 0;
		}
		else
		{
			im[S8(i)] = 1;
			S8i[S8(i)] = i;
		}
	}
	printf("Invertible\n");
	return 1;
}

int check_diff(int notinv)
{
	int max = 0;

	for (unsigned diff = 1; diff < 256; diff++)
	{
		uint8_t diff_table[256] = {0};

		if (notinv)
		{
			for (unsigned val = 0; val < 256; val++)
			{
				diff_table[S8(val) ^ S8(val ^ diff)] += 1;
			}
		}
		else
		{
			for (unsigned val = 0; val < 256; val++)
			{
				diff_table[S8inv(val) ^ S8inv(val ^ diff)] += 1;
			}
		}
		for (int i = 0; i < 256; i++)
			max = max > diff_table[i] ? max : diff_table[i];
	}

	printf("Max diff = %d\n", max);
	return max;
}

int check_diff4()
{
	int max = 0;

	for (unsigned diff = 1; diff < 16; diff++)
	{
		uint8_t diff_table[16] = {0};

		for (unsigned val = 0; val < 16; val++)
		{
			diff_table[S4(val) ^ S4(val ^ diff)] += 1;
		}
		for (int i = 0; i < 16; i++)
			max = max > diff_table[i] ? max : diff_table[i];
	}

	printf("Max diff = %d\n", max);
	return max;
}

int check_bn_diff(int notinv)
{
	int bn = 9;

	for (unsigned diff = 1; diff < 256; diff++)
	{
		for (unsigned val = 0; val < 256; val++)
		{
			int diffinw  = __builtin_popcount(diff & 0xff);
			int diffoutw;
			if (notinv)
			{
				diffoutw = __builtin_popcount((S8(val) ^ S8(val ^ diff)));
			}
			else
			{
				diffoutw = __builtin_popcount((S8inv(val) ^ S8inv(val ^ diff)));
			}

			bn = bn > diffinw + diffoutw ? diffinw + diffoutw : bn;
		}
	}

	printf("BN = %d\n", bn);
	return bn;

}

int check_lin(int notinv)
{
	int max = 0;

	for (unsigned maskin = 1; maskin < 256; maskin++)
	{
		int ones;
		int mones;

		for (unsigned maskout = 1; maskout < 256; maskout++)
		{
			ones = 0;
			mones = 0;
			if (notinv)
			{
				for (unsigned val = 0; val < 256; val++)
				{
					ones = Scalar8(val,maskin) ^ Scalar8(S8(val), maskout) ? ones : ones + 1;
					mones = Scalar8(val,maskin) ^ Scalar8(S8(val), maskout) ? mones + 1 : mones;
				}
			}
			else
			{
				for (unsigned val = 0; val < 256; val++)
				{
					ones = Scalar8(val,maskin) ^ Scalar8(S8inv(val), maskout) ? ones : ones + 1;
					mones = Scalar8(val,maskin) ^ Scalar8(S8inv(val), maskout) ? mones + 1 : mones;
				}
			}

			int sum = ones - mones;
//			int sum = 128 - ones;
			sum = sum >= 0 ? sum : -sum;
			max = max > sum ? max : sum;
		}
	}

	printf("Max lin = %d\n", max);
	return max;
}

int check_lin4()
{
	int max = 0;

	for (unsigned maskin = 1; maskin < 16; maskin++)
	{
		int ones;
		int mones;

		for (unsigned maskout = 1; maskout < 16; maskout++)
		{
			ones = 0;
			mones = 0;
				for (unsigned val = 0; val < 16; val++)
				{
					ones = Scalar4(val,maskin) ^ Scalar4(S4(val), maskout) ? ones : ones + 1;
					mones = Scalar4(val,maskin) ^ Scalar4(S4(val), maskout) ? mones + 1 : mones;
				}

			int sum = ones - mones;
//			int sum = 128 - ones;
			sum = sum >= 0 ? sum : -sum;
			max = max > sum ? max : sum;
		}
	}

	printf("Max lin = %d\n", max);
	return max;
}

int check_bn_lin(int notinv)
{
	int max = 0;

	for (unsigned maskin = 1; maskin < 256; maskin++)
	{
		int ones;
		int mones;

		if ((__builtin_popcount(maskin) & 0xff) == 1)
		{
			for (unsigned maskout = 1; maskout < 256; maskout++)
			{
				if ((__builtin_popcount(maskout) & 0xff) == 1)
				{
					ones = 0;
					mones = 0;
					if (notinv)
					{
						for (unsigned val = 0; val < 256; val++)
						{
							ones = Scalar8(val,maskin) ^ Scalar8(S8(val), maskout) ? ones : ones + 1;
							mones = Scalar8(val,maskin) ^ Scalar8(S8(val), maskout) ? mones + 1 : mones;
						}
					}
					else
					{
						for (unsigned val = 0; val < 256; val++)
						{
							ones = Scalar8(val,maskin) ^ Scalar8(S8inv(val), maskout) ? ones : ones + 1;
							mones = Scalar8(val,maskin) ^ Scalar8(S8inv(val), maskout) ? mones + 1 : mones;
						}
					}

					int sum = ones - mones;
					sum = sum >= 0 ? sum : -sum;
					if (sum > max)
					{
						printf("maskin = %02x, maskout = %02x, lin = %d\n", maskin, maskout, sum); 
					}
					max = max > sum ? max : sum;
				}
			}
		}
	}

	printf("Max lin weight 2 = %d\n", max);
	return max;

}

// Input: truth table of a Boolean function on lg(Wsize) + 6 bits
void moebius64(uint64_t *tab, uint64_t Wsize)
{
	uint64_t i, i0, i1;
	uint64_t step;

	for(step = 1; step < Wsize; step *= 2)
	{
		for(i1 = 0; i1 < Wsize; i1 += 2 * step)
		{
			for(i0 = 0; i0 < step; i0++)
			{
				i = i1 + i0;
				tab[i + step] ^= tab[i];
			}
		}
	}

	for(i = 0; i < Wsize; i++)
	{
		uint64_t tmp;
		tmp = tab[i];
		tmp ^= (tmp<<32);
		tmp ^= (tmp&0xffff0000ffffULL)<<16;
		tmp ^= (tmp&0xff00ff00ff00ffULL)<<8;
		tmp ^= (tmp&0xf0f0f0f0f0f0f0fULL)<<4;
		tmp ^= (tmp&0x3333333333333333ULL)<<2;
		tmp ^= (tmp&0x5555555555555555ULL)<<1;
		tab[i] = tmp;
	}
}

void project_boolean256(uint8_t ciph[256], int pos, uint64_t proji[4])
{
	for (int i = 0; i < 4; i++)
		proji[i] = 0;

	for (int i = 0; i < 256; i++)
	{
		int word = i / 64;
		int bpos = i % 64;
		uint8_t val = (ciph[i] >> pos) & 1;

		if (val)
			proji[word] |= (1ull << bpos);
	}
}


void project_boolean16(uint8_t ciph[16], int pos, uint64_t proji[1])
{
	proji[0] = 0;

	for (int i = 0; i < 16; i++)
	{
		uint8_t val = (ciph[i] >> pos) & 1;

		if (val)
			proji[0] |= (1ull << i);
	}
}

void get_anf8(uint8_t ciph[256], uint64_t bf[8][4])
{
	for (int i = 0; i < 8; i++)
	{
		project_boolean256(ciph, i, bf[i]);
		moebius64(bf[i], 4);
	}
}

void get_anf4(uint8_t ciph[16], uint64_t bf[4][1])
{
	for (int i = 0; i < 4; i++)
	{
		project_boolean16(ciph, i, bf[i]);
		moebius64(bf[i], 1);
	}
}


void check_degree(int notinv)
{
	uint8_t ciph[256];
	uint64_t bf[8][4];

	if (notinv)
	{
		for (int i = 0; i < 256; i++)
			ciph[i] = S8(i);
	}
	else
	{
		for (int i = 0; i < 256; i++)
			ciph[i] = S8inv(i);
	}

	get_anf8(ciph, bf);
	for (int bit = 0; bit < 8; bit++)
	{
//		int witp = 0;
//		for (int i = 0; i < (1 << 18); i++) witp += __builtin_popcountl(bf[bit][i]);
//		printf("Number of present terms on bit%d: %d\n", bit, witp);
		for (int deg = 0; deg < 8; deg++)
		{
			int degcount = 0;
			for (int poly = 0; poly < 256; poly++)
			{
				if (__builtin_popcount(poly) == deg)
				{
					int word = poly/64;
					int bt = poly%64;
					if (bf[bit][word] & (1ull << bt))
						degcount += 1;
				}
			}
			
			if ((deg > 0) && (deg < 8))
				printf("Number of present terms of deg %d on bit%d: %d\n", deg, bit, degcount);
		}
	}
}

void print_diff8(int notinv, FILE *stream)
{
	int one_diff[256] = {0};
	int inv_diff[256] = {0};

	fprintf(stream, "set terminal png nocrop enhanced font arial 20 size 1337,1337\n");
	if (notinv)
	{
		fprintf(stream, "set output '%s.png'\n", SNAMELD);
		fprintf(stream, "set title \"%s DDT in false colours\"\n", SNAMEC);
	}
	else
	{
		fprintf(stream, "set output '%s_inv.png'\n", SNAMELD);
//		fprintf(stream, "set title \"%s Inv DDT in false colours\"\n", SNAMEC);
	}
	fprintf(stream, "set palette rgbformula 21,22,23\n");
	//	fprintf(stream, "set cbrange [0:16]\n");
	fprintf(stream, "set xrange [0:257]\n");
	fprintf(stream, "set yrange [0:257]\n");
	fprintf(stream, "set clabel \"Hit count\"\n");
	fprintf(stream, "set xlabel \"Input difference\"\n");
	fprintf(stream, "set ylabel \"Output difference\"\n");
	fprintf(stream, "plot '-' using 2:1:3 with image\n");

	for (unsigned diff = 1; diff <= 0xff; diff++)
	{
		for (int i = 0; i < 256; i++)
		{
			one_diff[i] = 0;
			inv_diff[i] = 0;
		}
		for (unsigned val = 0; val < 256; val++)
		{
			one_diff[S8(val) ^ S8(val ^ diff)] += 1;
			inv_diff[S8inv(val) ^ S8inv(val ^ diff)] += 1;
		}
		fprintf(stream, "# Diff %02x\n", diff);
		for (int i = 0; i < 256; i++)
		{
			if (notinv)
			{
				fprintf(stream, "%d %d %d # %02x\n", diff, i+1, one_diff[i], i);
			}
			else
			{
				fprintf(stream, "%d %d %d # %02x\n", diff, i+1, inv_diff[i], i);
			}
		}
		fprintf(stream, "\n");
	}

	fprintf(stream, "e\n");

	return;
}

void print_lat8(int notinv, FILE *stream)
{
	fprintf(stream, "set terminal png nocrop enhanced font arial 20 size 1337,1337\n");
	if (notinv)
	{
		fprintf(stream, "set output '%s.png'\n", SNAMELL);
		fprintf(stream, "set title \"%s LAT in false colours\"\n", SNAMEC);
	}
	else
	{
		fprintf(stream, "set output '%s_inv.png'\n", SNAMELL);
//		fprintf(stream, "set title \"%s Inv LAT in false colours\"\n", SNAMEC);
	}
	fprintf(stream, "set palette rgbformula 21,22,23\n");
//	fprintf(stream, "set cbrange [0:64]\n");
	fprintf(stream, "set xrange [0:257]\n");
	fprintf(stream, "set yrange [0:257]\n");
	fprintf(stream, "set clabel \"Hit count\"\n");
	fprintf(stream, "set xlabel \"Input mask\"\n");
	fprintf(stream, "set ylabel \"Output mask\"\n");
	fprintf(stream, "plot '-' using 2:1:3 with image\n");

	for (unsigned maskin = 1; maskin < 256; maskin++)
	{
		int ones;
		int mones;

		for (unsigned maskout = 1; maskout < 256; maskout++)
		{
			ones = 0;
			mones = 0;
			if (notinv)
			{
				for (unsigned val = 0; val < 256; val++)
				{
					ones = Scalar8(val,maskin) ^ Scalar8(S8(val), maskout) ? ones : ones + 1;
					mones = Scalar8(val,maskin) ^ Scalar8(S8(val), maskout) ? mones + 1 : mones;
				}
			}
			else
			{
				for (unsigned val = 0; val < 256; val++)
				{
					ones = Scalar8(val,maskin) ^ Scalar8(S8inv(val), maskout) ? ones : ones + 1;
					mones = Scalar8(val,maskin) ^ Scalar8(S8inv(val), maskout) ? mones + 1 : mones;
				}
			}

			int sum = ones - mones;
			sum = sum >= 0 ? sum : -sum;
			fprintf(stream, "%d %d %d\n", maskin, maskout, sum);
		}
	}

	fprintf(stream, "\ne\n");

	return;
}

void print_anf8(int notinv, FILE *stream)
{
	fprintf(stream, "set terminal png nocrop enhanced font arial 10 size 1337,1337\n");
	if (notinv)
	{
		fprintf(stream, "set output '%s.png'\n", SNAMELA);
		fprintf(stream, "set title \"%s ANF\"\n", SNAMEC);
	}
	else
	{
		fprintf(stream, "set output '%s_inv.png'\n", SNAMELA);
		fprintf(stream, "set title \"%s Inv ANF\"\n", SNAMEC);
	}
	fprintf(stream, "set palette rgbformula 21,22,23\n");
	//	fprintf(stream, "set cbrange [0:16]\n");
	fprintf(stream, "set xrange [0:256]\n");
	fprintf(stream, "set yrange [0:9]\n");
	fprintf(stream, "set xlabel \"Poly\"\n");
	fprintf(stream, "set ylabel \"Output bit\"\n");
	fprintf(stream, "plot '-' using 2:1:3 with image\n");
	
	uint8_t ciph[256];
	uint64_t bf[8][4];

	if (notinv)
	{
		for (int i = 0; i < 256; i++)
			ciph[i] = S8(i);
	}
	else
	{
		for (int i = 0; i < 256; i++)
			ciph[i] = S8inv(i);
	}

	get_anf8(ciph, bf);
	for (int bit = 0; bit < 8; bit++)
	{
		int polyord = 0;
//		for (int deg = 0; deg < 8; deg++)
//		{
			for (int poly = 0; poly < 256; poly++)
			{
//				if (__builtin_popcount(poly) == deg)
//				{
					int word = poly/64;
					int bt = poly%64;
					if (bf[bit][word] & (1ull << bt))
					{
						fprintf(stream, "%d %d %d\n", bit + 1, poly, 1);
					}
					else
					{
						fprintf(stream, "%d %d %d\n", bit + 1, poly, 0);
					}
//				}
//			}
		}
	}

	fprintf(stream, "\ne\n");

	return;
}


void print_diff4(int notinv, FILE *stream)
{
	int one_diff[16] = {0};

	fprintf(stream, "set terminal png nocrop enhanced font arial 20 size 1337,1337\n");
	if (notinv)
	{
		fprintf(stream, "set output '%s.png'\n", S4NAMELD);
		fprintf(stream, "set title \"%s DDT in false colours\"\n", S4NAMEC);
	}
	else
	{
		fprintf(stream, "set output '%s_inv.png'\n", S4NAMELD);
//		fprintf(stream, "set title \"%s Inv DDT in false colours\"\n", SNAMEC);
	}
	fprintf(stream, "set palette rgbformula 21,22,23\n");
	//	fprintf(stream, "set cbrange [0:16]\n");
	fprintf(stream, "set xrange [0:17]\n");
	fprintf(stream, "set yrange [0:17]\n");
	fprintf(stream, "set clabel \"Hit count\"\n");
	fprintf(stream, "set xlabel \"Input difference\"\n");
	fprintf(stream, "set ylabel \"Output difference\"\n");
	fprintf(stream, "plot '-' using 2:1:3 with image\n");

	for (unsigned diff = 1; diff <= 0xf; diff++)
	{
		for (int i = 0; i < 16; i++)
		{
			one_diff[i] = 0;
		}
		for (unsigned val = 0; val < 16; val++)
		{
			one_diff[S4(val) ^ S4(val ^ diff)] += 1;
		}
		fprintf(stream, "# Diff %02x\n", diff);
		for (int i = 0; i < 16; i++)
		{
			if (notinv)
			{
				fprintf(stream, "%d %d %d # %02x\n", diff, i+1, one_diff[i], i);
			}
		}
		fprintf(stream, "\n");
	}

	fprintf(stream, "e\n");

	return;
}

void print_lat4(int notinv, FILE *stream)
{
	fprintf(stream, "set terminal png nocrop enhanced font arial 20 size 1337,1337\n");
	if (notinv)
	{
		fprintf(stream, "set output '%s.png'\n", S4NAMELL);
		fprintf(stream, "set title \"%s LAT in false colours\"\n", S4NAMEC);
	}
	else
	{
		fprintf(stream, "set output '%s_inv.png'\n", S4NAMELL);
//		fprintf(stream, "set title \"%s Inv LAT in false colours\"\n", SNAMEC);
	}
	fprintf(stream, "set palette rgbformula 21,22,23\n");
//	fprintf(stream, "set cbrange [0:64]\n");
	fprintf(stream, "set xrange [0:17]\n");
	fprintf(stream, "set yrange [0:17]\n");
	fprintf(stream, "set clabel \"Hit count\"\n");
	fprintf(stream, "set xlabel \"Input mask\"\n");
	fprintf(stream, "set ylabel \"Output mask\"\n");
	fprintf(stream, "plot '-' using 2:1:3 with image\n");

	for (unsigned maskin = 1; maskin < 16; maskin++)
	{
		int ones;
		int mones;

		for (unsigned maskout = 1; maskout < 16; maskout++)
		{
			ones = 0;
			mones = 0;
			if (notinv)
			{
				for (unsigned val = 0; val < 16; val++)
				{
					ones = Scalar4(val,maskin) ^ Scalar4(S4(val), maskout) ? ones : ones + 1;
					mones = Scalar4(val,maskin) ^ Scalar4(S4(val), maskout) ? mones + 1 : mones;
				}
			}

			int sum = ones - mones;
			sum = sum >= 0 ? sum : -sum;
			fprintf(stream, "%d %d %d\n", maskin, maskout, sum);
		}
	}

	fprintf(stream, "\ne\n");

	return;
}

void print_anf4(int notinv, FILE *stream)
{
	fprintf(stream, "set terminal png nocrop enhanced font arial 10 size 1337,1337\n");
	if (notinv)
	{
		fprintf(stream, "set output '%s.png'\n", S4NAMELA);
		fprintf(stream, "set title \"%s ANF\"\n", S4NAMEC);
	}
	else
	{
		fprintf(stream, "set output '%s_inv.png'\n", S4NAMELA);
		fprintf(stream, "set title \"%s Inv ANF\"\n", S4NAMEC);
	}
	fprintf(stream, "set palette rgbformula 21,22,23\n");
	//	fprintf(stream, "set cbrange [0:16]\n");
	fprintf(stream, "set xrange [0:16]\n");
	fprintf(stream, "set yrange [0:5]\n");
	fprintf(stream, "set xlabel \"Poly\"\n");
	fprintf(stream, "set ylabel \"Output bit\"\n");
	fprintf(stream, "plot '-' using 2:1:3 with image\n");
	
	uint8_t ciph[16];
	uint64_t bf[4][1];

	if (notinv)
	{
		for (int i = 0; i < 16; i++)
			ciph[i] = S4(i);
	}

	get_anf4(ciph, bf);
	for (int bit = 0; bit < 4; bit++)
	{
		int polyord = 0;
//		for (int deg = 0; deg < 8; deg++)
//		{
			for (int poly = 0; poly < 16; poly++)
			{
//				if (__builtin_popcount(poly) == deg)
//				{
					if (bf[bit][poly] & (1ull << poly))
					{
						fprintf(stream, "%d %d %d\n", bit + 1, poly, 1);
					}
					else
					{
						fprintf(stream, "%d %d %d\n", bit + 1, poly, 0);
					}
//				}
//			}
		}
	}

	fprintf(stream, "\ne\n");

	return;
}

void check_fp(void)
{
	int count = 0;
	for (int i = 0; i < 256; i++)
	{
		if (S8(i) == i)
		{
		   printf("%d (%x)\n", i, i);
		   count += 1;
		}
	}
	printf("There's %d fixed points\n", count);

	return;
}

int main()
{
	check_invertible();
//	check_diff(1);
//	check_bn_diff(1);
//	check_lin(1);
//	check_bn_lin(1);
//	check_degree(1);
//	printf("Properties of the inverse\n");
//	check_diff(0);
//	check_bn_diff(0);
//	check_lin(0);
//	check_bn_lin(0);
//	check_degree(0);
//	check_fp();
//	print_diff8(1, stderr);
//	print_lat8(1, stderr);
//	print_anf8(1, stderr);
//	print_diff4(1, stderr);
//	print_lat4(1, stderr);
//	print_anf4(1, stderr);
	check_diff4();
	check_lin4();

	return 0;
}
